module Test_RS

include(joinpath(@__DIR__, "..", "..", "src", "ZMPBipedRobot.jl"))
import .ZMPBipedRobot
const ZMProbot = ZMPBipedRobot

using RigidBodyDynamics
using Test

sleep(0.1) # used for good printing
println("Started test")
Δt = 1e-3
Kp = 10000.0
Kd = 100.0
Ki = 100.0
ctrl = true
@testset "Robot Simulator" begin
    # Construct the biped robot which store the geomtrical propreties and the path wanted 
    URDFfileName = "ZMP_2DBipedRobot.urdf"
    br = ZMProbot.BipedRobot(;
        readFile = true,
        URDFfileName = URDFfileName,
        paramFileName = "param_test.jl",
    )

    # Construct the Preview Controller
    pc = ZMProbot.PreviewController(; br = br)

    # Run the Foot Planer Algorithm and get the foot position 
    fp = ZMProbot.FootPlanner(; br = br)

    # Get the ZMP reference trajectory 
    zt = ZMProbot.ZMPTrajectory(; br = br, fp = fp)

    # Convert the ZMP reference trajectory into CoM trajectory
    ct = ZMProbot.CoMTrajectory(; br = br, pc = pc, zt = zt)

    # Get the Swing Foot trajectory 
    sf = ZMProbot.SwingFootTrajectory(; br = br, fp = fp, zt = zt)

    # Get the joint trajectory from the all path 
    ik = ZMProbot.InverseKinematics(; br = br, fp = fp, ct = ct, sf = sf)

    # Store into more convienant variables 
    qr = ik.q_r
    ql = ik.q_l
    qref = [ql[:, 1] qr[:, 1] ql[:, 2] qr[:, 2]]

    CoM = reduce(hcat, ct.CoM)
    ZMP = reduce(hcat, zt.ZMP)
    tplot = reduce(vcat, zt.timeVec)
    tend = 0.05      # Simulation time 

    # Construct the robot in the simulation engine 
    rs = ZMProbot.RobotSimulator(;
        fileName = URDFfileName,
        symbolic = false,
        add_contact_points = true,
        add_gravity = true,
        add_flat_ground = true,
    )
    # Generate the visualiser 
    vis = ZMProbot.set_visulalizer(; mechanism = rs.mechanism)

    # Intiial configuration 
    boom = [0, 0]
    actuators = [0, 0, 0, 0]
    foot = [0, 0]
    ZMProbot.set_nominal!(rs, vis, boom, actuators, foot)
    ZMProbot.set_initialbody!(rs, vis)

    # Simulate the robot 
    controller! = ZMProbot.trajectory_controller!(rs, tplot, qref, Δt, Kp, Ki, Kd, ctrl)
    ts, qs, vs = RigidBodyDynamics.simulate(rs.state, tend, controller!; Δt = Δt)

    torque_sim = rs.torques

    @test qs == [
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        [
            -1.3031753962641616e-5,
            -4.90560721318855e-6,
            -0.0001254642605159546,
            -0.0001254642605159546,
            0.00020354346889521482,
            0.00020354346889521482,
            -9.066929603064182e-5,
            -9.066929603064184e-5,
        ],
        [
            -5.8053358593172674e-5,
            -1.963315826329752e-5,
            -0.0005644157660900375,
            -0.0005644157660900375,
            0.0009308134968324932,
            0.0009308134968324933,
            -0.0004445229997889594,
            -0.0004445229997889594,
        ],
        [
            -0.00012443811818898016,
            -4.417649946910298e-5,
            -0.0012221208053863116,
            -0.0012221208053863116,
            0.002048200391075468,
            0.002048200391075468,
            -0.0010415866972180061,
            -0.0010415866972180061,
        ],
        [
            -0.00020405862128678853,
            -7.845304307152831e-5,
            -0.0020249226223388044,
            -0.0020249226223388044,
            0.003446272677417347,
            0.0034462726774173464,
            -0.0018497444187204428,
            -0.0018497444187204428,
        ],
        [
            -0.00029071289646097093,
            -0.00012225187711494272,
            -0.002915426067114354,
            -0.002915426067114354,
            0.00503308196173175,
            0.00503308196173175,
            -0.0028197623814927176,
            -0.0028197623814927176,
        ],
        [
            -0.00037969921125772214,
            -0.00017519428089739867,
            -0.003848981222421601,
            -0.003848981222421601,
            0.006729298509805254,
            0.006729298509805254,
            -0.0038839620410430153,
            -0.0038839620410430157,
        ],
        [
            -0.0004674618238060542,
            -0.00023670304541093754,
            -0.004790814254886469,
            -0.004790814254886469,
            0.008464478776404686,
            0.008464478776404686,
            -0.004962815617465347,
            -0.004962815617465347,
        ],
        [
            -0.0005512991298821158,
            -0.0003059800339383982,
            -0.005713727768345381,
            -0.005713727768345381,
            0.010174406032934392,
            0.010174406032934392,
            -0.00598088016379807,
            -0.00598088016379807,
        ],
        [
            -0.00062912630848463,
            -0.00038199192894676373,
            -0.006596287044708176,
            -0.006596287044708176,
            0.011799321170818422,
            0.011799321170818422,
            -0.006887707854575492,
            -0.006887707854575493,
        ],
        [
            -0.0006992842840562105,
            -0.0004634643559310041,
            -0.007421399018911842,
            -0.007421399018911843,
            0.013282744742602732,
            0.013282744742602734,
            -0.007672796576080676,
            -0.007672796576080678,
        ],
        [
            -0.0007604898450159704,
            -0.0005488909421003838,
            -0.008176999580674287,
            -0.008176999580674289,
            0.014584713243194302,
            0.014584713243194304,
            -0.008315503542642062,
            -0.008315503542642062,
        ],
        [
            -0.0008125883998625395,
            -0.0006366008846299168,
            -0.008869660133600143,
            -0.008869660133600145,
            0.015793129058714047,
            0.01579312905871405,
            -0.008460685504529903,
            -0.008460685504529903,
        ],
        [
            -0.0008556390524327964,
            -0.0007247314861222529,
            -0.009499536400186082,
            -0.009499536400186083,
            0.016906785557962267,
            0.01690678555796227,
            -0.0083033080135963,
            -0.008303308013596298,
        ],
        [
            -0.0008901366688414356,
            -0.0008112965071939257,
            -0.010072614709184476,
            -0.010072614709184477,
            0.017957865495506885,
            0.01795786549550689,
            -0.00806972144305872,
            -0.008069721443058717,
        ],
        [
            -0.0009167995784081914,
            -0.0008942291405232443,
            -0.01059598009341743,
            -0.010595980093417431,
            0.018977235151885062,
            0.018977235151885066,
            -0.008124582929438668,
            -0.008124582929438672,
        ],
        [
            -0.0009364092585753686,
            -0.0009714495147507556,
            -0.011074738847057309,
            -0.011074738847057309,
            0.019968877008031283,
            0.019968877008031283,
            -0.008665319243996352,
            -0.008665319243996371,
        ],
        [
            -0.0009499264701599644,
            -0.0010409743587657663,
            -0.01151766636840041,
            -0.01151766636840041,
            0.020968483360316502,
            0.020968483360316502,
            -0.009545022865214494,
            -0.009545022865214517,
        ],
        [
            -0.0009579869441634074,
            -0.0011009460906791506,
            -0.01192541046248342,
            -0.01192541046248342,
            0.021942145513996245,
            0.021942145513996245,
            -0.010608142097710339,
            -0.010608142097710355,
        ],
        [
            -0.0009611197621260689,
            -0.0011497494038495016,
            -0.012300153546604945,
            -0.012300153546604945,
            0.022878715938495272,
            0.022878715938495272,
            -0.01150916549268348,
            -0.011509165492683479,
        ],
        [
            -0.0009596421822493318,
            -0.0011860697068937847,
            -0.012642121673201034,
            -0.012642121673201034,
            0.02375842761126249,
            0.023758427611262487,
            -0.012123015440294668,
            -0.012123015440294649,
        ],
        [
            -0.0009727376703240629,
            -0.0012093218265302745,
            -0.01313551865899734,
            -0.01313551865899734,
            0.024879734879782768,
            0.02487973487978276,
            -0.012530675185441929,
            -0.012530675185441894,
        ],
        [
            -0.0010112748957157764,
            -0.001219241587349447,
            -0.013889194591040382,
            -0.013889194591040382,
            0.026413394912585996,
            0.026413394912585986,
            -0.013165406918059333,
            -0.013165406918059298,
        ],
        [
            -0.0010654956265549577,
            -0.0012156524217649073,
            -0.01481653527581175,
            -0.01481653527581175,
            0.028241181157376605,
            0.028241181157376595,
            -0.014018476771400232,
            -0.014018476771400211,
        ],
        [
            -0.0011278143938696207,
            -0.0011988153058322124,
            -0.01584839596807012,
            -0.01584839596807012,
            0.03025438979441874,
            0.030254389794418732,
            -0.015190045986032541,
            -0.015190045986032526,
        ],
        [
            -0.0011926056402503675,
            -0.001169405967816432,
            -0.01693315024272258,
            -0.016933150242722583,
            0.03237374760456838,
            0.032373747604568376,
            -0.016429659570154357,
            -0.016429659570154343,
        ],
        [
            -0.0012553990850702043,
            -0.0011284395651969313,
            -0.018027264266584235,
            -0.01802726426658424,
            0.03450594333145283,
            0.03450594333145282,
            -0.017769367249206823,
            -0.017769367249206823,
        ],
        [
            -0.0013130569807383078,
            -0.001077257685384958,
            -0.01910145940969525,
            -0.01910145940969525,
            0.036602860649805766,
            0.03660286064980575,
            -0.018906576675449926,
            -0.01890657667544994,
        ],
        [
            -0.0013633224464079548,
            -0.0010174614305971613,
            -0.020134995632260816,
            -0.020134995632260816,
            0.03863781495814146,
            0.03863781495814145,
            -0.019884427356149,
            -0.019884427356149006,
        ],
        [
            -0.0014048166496101822,
            -0.0009508208658971635,
            -0.021113965963115648,
            -0.021113965963115648,
            0.04058592894730436,
            0.040585928947304346,
            -0.020642106097202007,
            -0.020642106097201986,
        ],
        [
            -0.0014366751924561592,
            -0.0008792229443744061,
            -0.0220298056834615,
            -0.0220298056834615,
            0.042439422886563175,
            0.042439422886563154,
            -0.021390890815125282,
            -0.021390890815125255,
        ],
        [
            -0.0014585866894189614,
            -0.0008045783013301726,
            -0.022877785937952252,
            -0.02287778593795225,
            0.04418611242442927,
            0.04418611242442924,
            -0.02214783724759728,
            -0.02214783724759726,
        ],
        [
            -0.0014706725854169566,
            -0.0007288052272516016,
            -0.023659355867191322,
            -0.02365935586719132,
            0.04585126350464065,
            0.045851263504640624,
            -0.022957641594071185,
            -0.022957641594071158,
        ],
        [
            -0.0014733246008384931,
            -0.0006537029586301188,
            -0.02437511383763543,
            -0.024375113837635427,
            0.047426902412337386,
            0.047426902412337366,
            -0.023793331375405007,
            -0.023793331375404965,
        ],
        [
            -0.0014670277383361395,
            -0.0005809505185675861,
            -0.02502743556695402,
            -0.025027435566954016,
            0.04891683754133048,
            0.04891683754133046,
            -0.02468073549381786,
            -0.024680735493817815,
        ],
        [
            -0.0014523662994180204,
            -0.0005120449916805622,
            -0.025618491677003516,
            -0.025618491677003512,
            0.050311435044327106,
            0.050311435044327085,
            -0.02556885310271123,
            -0.025568853102711203,
        ],
        [
            -0.0014301097349192816,
            -0.0004483423418607952,
            -0.026155034706154486,
            -0.026155034706154483,
            0.051640401963440274,
            0.051640401963440254,
            -0.02636352268277749,
            -0.026363522682777484,
        ],
        [
            -0.001400811858736688,
            -0.0003909199393270745,
            -0.026637020292599314,
            -0.02663702029259931,
            0.05286786675464796,
            0.05286786675464794,
            -0.02715281406089575,
            -0.027152814060895758,
        ],
        [
            -0.001365348267880273,
            -0.0003407628281121599,
            -0.027074046109634604,
            -0.027074046109634604,
            0.054045625461556564,
            0.05404562546155654,
            -0.027794992535224144,
            -0.027794992535224158,
        ],
        [
            -0.0013243139333521644,
            -0.00029856013243257577,
            -0.027466660060872568,
            -0.027466660060872568,
            0.055140956514766915,
            0.055140956514766894,
            -0.02842274703048653,
            -0.028422747030486546,
        ],
        [
            -0.0012783758394834897,
            -0.0002648864802040822,
            -0.02782041436250918,
            -0.02782041436250918,
            0.056171789276000604,
            0.05617178927600058,
            -0.029018969092557093,
            -0.0290189690925571,
        ],
        [
            -0.001246784076619719,
            -0.00024092464187011224,
            -0.028319601188948364,
            -0.028319601188948364,
            0.05744224627300409,
            0.057442246273004066,
            -0.02970639685560444,
            -0.02970639685560444,
        ],
        [
            -0.0012408239746877206,
            -0.00022750197224658843,
            -0.029077662172582484,
            -0.029077662172582484,
            0.0591552843572964,
            0.05915528435729638,
            -0.030603181437316728,
            -0.030603181437316718,
        ],
        [
            -0.0012512610984352485,
            -0.00022452338009257673,
            -0.030010602013487788,
            -0.030010602013487788,
            0.06119204752717327,
            0.06119204752717325,
            -0.0316865857955795,
            -0.03168658579557948,
        ],
        [
            -0.0012710965366528183,
            -0.00023186763993685245,
            -0.031053307662295783,
            -0.031053307662295783,
            0.06345528645626783,
            0.0634552864562678,
            -0.03292699029889344,
            -0.032926990298893406,
        ],
        [
            -0.0012951029050708054,
            -0.0002493853469220439,
            -0.03215576159178693,
            -0.03215576159178693,
            0.06586543492557112,
            0.06586543492557109,
            -0.03428799874946771,
            -0.03428799874946766,
        ],
        [
            -0.0013194454947874188,
            -0.00027689626913308523,
            -0.03327994119383948,
            -0.03327994119383948,
            0.06835724699943489,
            0.06835724699943486,
            -0.035726894898999245,
            -0.0357268948989992,
        ],
        [
            -0.0013413722982299565,
            -0.0003141777916013038,
            -0.03439726917951362,
            -0.03439726917951362,
            0.0708767673740509,
            0.07087676737405087,
            -0.037196914679798736,
            -0.03719691467979869,
        ],
        [
            -0.0013589592453343746,
            -0.000360942777619149,
            -0.03548651397474428,
            -0.03548651397474428,
            0.0733785881355674,
            0.07337858813556739,
            -0.03865314085293536,
            -0.038653140852935315,
        ],
        [
            -0.0013708999202166965,
            -0.00041681004236255535,
            -0.03653206578005892,
            -0.03653206578005892,
            0.07582343839632212,
            0.07582343839632211,
            -0.04006189763827313,
            -0.04006189763827309,
        ],
        [
            -0.001376331356189253,
            -0.00048127131150882776,
            -0.03752252592311488,
            -0.03752252592311488,
            0.0781761149410119,
            0.07817611494101188,
            -0.041410769187714656,
            -0.041410769187714636,
        ],
    ]

    @test vs == [
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        [
            -0.031045025021418492,
            -0.009814418125110452,
            -0.30008216730419274,
            -0.30008216730419274,
            0.49013723345560833,
            0.4901372334556084,
            -0.2248142857409521,
            -0.22481428574095214,
        ],
        [
            -0.05723686498799687,
            -0.019640752718480536,
            -0.562115373772223,
            -0.5621153737722231,
            0.9422511880627431,
            0.9422511880627432,
            -0.4796019147144,
            -0.47960191471439995,
        ],
        [
            -0.07417698086259863,
            -0.029432878108330222,
            -0.7410097266572631,
            -0.7410097266572631,
            1.2742862424457735,
            1.2742862424457735,
            -0.709240604032614,
            -0.709240604032614,
        ],
        [
            -0.08402954946532638,
            -0.03908564000695116,
            -0.8550211724126475,
            -0.8550211724126474,
            1.5065361550530858,
            1.5065361550530858,
            -0.8989168775624089,
            -0.898916877562409,
        ],
        [
            -0.08849481755288008,
            -0.048449399524263775,
            -0.9185388310297494,
            -0.9185388310297493,
            1.6538625908682187,
            1.6538625908682187,
            -1.0298372793880697,
            -1.02983727938807,
        ],
        [
            -0.08888484327534668,
            -0.05733962487692905,
            -0.9427720052912659,
            -0.9427720052912659,
            1.7268305661762546,
            1.7268305661762549,
            -1.085250564463779,
            -1.0852505644637793,
        ],
        [
            -0.08618995760043989,
            -0.06554523761908375,
            -0.9363586200140387,
            -0.9363586200140387,
            1.7328272952437174,
            1.7328272952437176,
            -1.0598062429396105,
            -1.0598062429396107,
        ],
        [
            -0.0811372000276008,
            -0.07283681764551078,
            -0.9058947090369951,
            -0.9058947090369951,
            1.677068049365847,
            1.6770680493658472,
            -0.9678858259657563,
            -0.9678858259657566,
        ],
        [
            -0.07424082171879645,
            -0.07897480163956185,
            -0.8563749876007974,
            -0.8563749876007976,
            1.5633504767215622,
            1.5633504767215625,
            -0.8440896660785763,
            -0.8440896660785765,
        ],
        [
            -0.06584436412566919,
            -0.08371821680213003,
            -0.7915381582968379,
            -0.791538158296838,
            1.3944843225426873,
            1.3944843225426875,
            -0.7304451297208094,
            -0.7304451297208095,
        ],
        [
            -0.05664794583287686,
            -0.08687343005819347,
            -0.7264790492121127,
            -0.7264790492121129,
            1.2807146097183282,
            1.2807146097183288,
            -0.32408395838924425,
            -0.3240839583892432,
        ],
        [
            -0.04765482574914659,
            -0.08826691475043384,
            -0.6677440016336491,
            -0.6677440016336491,
            1.2282257066073312,
            1.228225706607331,
            0.24487275331292352,
            0.24487275331292185,
        ],
        [
            -0.03860291199431799,
            -0.0876999102365663,
            -0.6032581586530733,
            -0.6032581586530734,
            1.1128568913959802,
            1.1128568913959822,
            0.3293073862145162,
            0.32930738621451955,
        ],
        [
            -0.030532086635570038,
            -0.08513418766563106,
            -0.556128490457568,
            -0.556128490457568,
            1.1263238126040056,
            1.1263238126040074,
            0.38946225156117864,
            0.38946225156117786,
        ],
        [
            -0.022950655913665657,
            -0.08044623946768334,
            -0.505964202298918,
            -0.5059642022989176,
            1.0714213219652047,
            1.0714213219652018,
            -0.13394269532701272,
            -0.13394269532702907,
        ],
        [
            -0.01639280759262783,
            -0.07373804908592166,
            -0.46868906888879586,
            -0.4686890688887954,
            1.092890133869424,
            1.092890133869421,
            -0.5165898094846744,
            -0.5165898094846888,
        ],
        [
            -0.01073820967822181,
            -0.06509641172004618,
            -0.43599777888152996,
            -0.4359977788815299,
            1.107947643793937,
            1.1079476437939364,
            -0.7351230597836584,
            -0.7351230597836536,
        ],
        [
            -0.005473046710394957,
            -0.054684369694600304,
            -0.39999201419421665,
            -0.3999920141942165,
            1.0595131022151911,
            1.0595131022151907,
            -0.8002074563643349,
            -0.8002074563643263,
        ],
        [
            -0.0008565060758691616,
            -0.04282575382764346,
            -0.37135168899513055,
            -0.3713516889951304,
            1.050636027038134,
            1.0506360270381334,
            -0.42446736968082577,
            -0.42446736968080245,
        ],
        [
            0.003732745469905994,
            -0.029789316530914725,
            -0.33570007454659107,
            -0.33570007454659073,
            0.9582906072670974,
            0.9582906072670948,
            -0.1892808994909539,
            -0.1892808994909406,
        ],
        [
            -0.027833235175800084,
            -0.016749931565124852,
            -0.6559240659516061,
            -0.655924065951606,
            1.5170532044711922,
            1.5170532044711915,
            -0.06629358890877704,
            -0.06629358890876734,
        ],
        [
            -0.04767555737398648,
            -0.0032082229038759213,
            -0.8608656822972658,
            -0.8608656822972657,
            1.7902165474215588,
            1.790216547421558,
            -0.5414484385542895,
            -0.5414484385543002,
        ],
        [
            -0.05955130004471532,
            0.0101804684827771,
            -1.0062075835124138,
            -1.0062075835124136,
            2.1087618512341697,
            2.1087618512341684,
            -0.5749760608527903,
            -0.5749760608528036,
        ],
        [
            -0.06422889542580601,
            0.023218060052337244,
            -1.0722179704226587,
            -1.0722179704226589,
            2.156733016281346,
            2.1567330162813496,
            -1.0724052863583875,
            -1.0724052863583886,
        ],
        [
            -0.06469937758472662,
            0.03525396421195491,
            -1.1129454261445852,
            -1.1129454261445852,
            2.3146396313744866,
            2.3146396313744897,
            -0.8331672557531701,
            -0.8331672557531755,
        ],
        [
            -0.060500839141842844,
            0.046288364564768064,
            -1.0917593579105134,
            -1.0917593579105127,
            2.1690770055890667,
            2.169077005589063,
            -1.1746212670054388,
            -1.1746212670054588,
        ],
        [
            -0.05452937370157947,
            0.05563766371731232,
            -1.0726516583362053,
            -1.0726516583362047,
            2.2307009019236594,
            2.2307009019236568,
            -0.6149252233277209,
            -0.6149252233277284,
        ],
        [
            -0.04590652936019546,
            0.06350187103888161,
            -1.010162271565405,
            -1.010162271565405,
            2.0261866287760846,
            2.0261866287760855,
            -0.8007489141584068,
            -0.8007489141583838,
        ],
        [
            -0.03702799992406019,
            0.06931091710713604,
            -0.9621731748694348,
            -0.9621731748694347,
            2.0388295093651205,
            2.0388295093651214,
            -0.3463726286819596,
            -0.3463726286819326,
        ],
        [
            -0.02677095580732668,
            0.07342966199452926,
            -0.8829560014424955,
            -0.8829560014424949,
            1.815170759460934,
            1.8151707594609292,
            -0.727884979537282,
            -0.7278849795372886,
        ],
        [
            -0.017132324954301576,
            0.07541820629365324,
            -0.8246522077816759,
            -0.8246522077816753,
            1.805405480243401,
            1.8054054802433965,
            -0.49758240982411983,
            -0.497582409824126,
        ],
        [
            -0.007196490386412257,
            0.07571996361161351,
            -0.7488372832258269,
            -0.748837283225827,
            1.6312120045444904,
            1.631212004544494,
            -0.8127654341467896,
            -0.8127654341467724,
        ],
        [
            0.0017539224616214657,
            0.07411127264861303,
            -0.6912331275990712,
            -0.6912331275990713,
            1.607894803589225,
            1.6078948035892284,
            -0.6472202951025623,
            -0.6472202951025514,
        ],
        [
            0.010667117615506516,
            0.07107211067070271,
            -0.6202352892137115,
            -0.6202352892137113,
            1.437148409399255,
            1.4371484093992546,
            -0.9217864150264548,
            -0.9217864150264663,
        ],
        [
            0.018494162670543768,
            0.0664554147574997,
            -0.5676830612243853,
            -0.5676830612243852,
            1.407056373469221,
            1.407056373469221,
            -0.7125715597279764,
            -0.7125715597279987,
        ],
        [
            0.025960172568918232,
            0.06075558513563284,
            -0.506503866406313,
            -0.5065038664063131,
            1.2590263694283206,
            1.2590263694283215,
            -0.8562690865356224,
            -0.8562690865356402,
        ],
        [
            0.032473127340959605,
            0.05389814161813208,
            -0.4611230269813279,
            -0.46112302698132795,
            1.2264815634640245,
            1.2264815634640251,
            -0.6428611074783731,
            -0.6428611074783827,
        ],
        [
            0.03835649469920781,
            0.046290275115985186,
            -0.41386782381810416,
            -0.4138678238181042,
            1.1330539751439954,
            1.1330539751439959,
            -0.6383726470477787,
            -0.6383726470477803,
        ],
        [
            0.04360093739308091,
            0.038019750444696465,
            -0.37228379306562726,
            -0.3722837930656273,
            1.0605792521947963,
            1.0605792521947965,
            -0.6143110125801313,
            -0.6143110125801269,
        ],
        [
            0.04816135142087391,
            0.02925899270041516,
            -0.3360760626438589,
            -0.336076062643859,
            1.0031450143305531,
            1.0031450143305536,
            -0.5767657756243771,
            -0.5767657756243693,
        ],
        [
            0.01703007482931,
            0.018678439771772106,
            -0.6443629009891281,
            -0.6443629009891282,
            1.5135206161461487,
            1.5135206161461494,
            -0.7953087171421455,
            -0.7953087171421364,
        ],
        [
            -0.0035692984837696687,
            0.008182919759769484,
            -0.857746010343101,
            -0.8577460103431013,
            1.8927177216315385,
            1.8927177216315392,
            -0.9944491653729782,
            -0.994449165372969,
        ],
        [
            -0.016137141882224,
            -0.0022052805253136385,
            -0.997269578187152,
            -0.997269578187152,
            2.1645771450078044,
            2.1645771450078053,
            -1.16741297900657,
            -1.1674129790065597,
        ],
        [
            -0.02266135383967325,
            -0.012458160695730702,
            -1.0797955113101199,
            -1.0797955113101199,
            2.348623653820841,
            2.348623653820842,
            -1.3073384318699282,
            -1.3073384318699188,
        ],
        [
            -0.024711296895311968,
            -0.022547217516186636,
            -1.118767276057912,
            -1.118767276057912,
            2.460790992951611,
            2.4607909929516123,
            -1.4075612257073737,
            -1.407561225707366,
        ],
        [
            -0.023514441276644597,
            -0.032437718022769395,
            -1.1248229200361854,
            -1.1248229200361854,
            2.513839161260735,
            2.5138391612607367,
            -1.4624055990054838,
            -1.4624055990054807,
        ],
        [
            -0.02001791119940941,
            -0.04207779527823781,
            -1.1062899340961518,
            -1.1062899340961518,
            2.5176361148206845,
            2.5176361148206863,
            -1.4700819809924834,
            -1.4700819809924843,
        ],
        [
            -0.014937862931383765,
            -0.051389043883158354,
            -1.0695949178275546,
            -1.0695949178275546,
            2.47946500390183,
            2.479465003901832,
            -1.4366779264109342,
            -1.4366779264109384,
        ],
        [
            -0.008799460871459466,
            -0.06026120509180871,
            -1.0196071254294388,
            -1.0196071254294388,
            2.404374581772513,
            2.4043745817725153,
            -1.378708731292705,
            -1.3787087312927162,
        ],
        [
            -0.0019692846643315333,
            -0.06855028963321289,
            -0.959922053495068,
            -0.959922053495068,
            2.295497859719827,
            2.2954978597198292,
            -1.3213472208476968,
            -1.3213472208477108,
        ],
    ]

    @test torque_sim == Array[
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        [
            0.0,
            0.0,
            -0.9978785337672311,
            -0.9978785337672311,
            -0.08290756424715556,
            -0.08290756424715556,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            -0.5332382937520359,
            -0.5332382937520359,
            0.0038420653271056815,
            0.0038420653271056815,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            -0.2907941568412171,
            -0.29079415684121657,
            0.044859324203731385,
            0.04485932420373147,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            -0.11048474183712259,
            -0.11048474183712259,
            0.07230121394709482,
            0.07230121394709482,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            0.02195442887192328,
            0.021954428871922893,
            0.08993478878227278,
            0.08993478878227268,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            0.11876364497294721,
            0.11876364497294721,
            0.10114738194948718,
            0.10114738194948718,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            0.190384252240655,
            0.19038425224065494,
            0.1089668163769648,
            0.10896681637696479,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            0.24561656339975907,
            0.24561656339975887,
            0.1160570026372218,
            0.11605700263722171,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            0.29176524285134775,
            0.29176524285134775,
            0.12471495512772746,
            0.12471495512772746,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            0.3348249493294786,
            0.334824949329479,
            0.13689588718551518,
            0.13689588718551532,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            0.3573213456786542,
            0.35732134567865426,
            0.14557318015674275,
            0.14557318015674275,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            0.40721626785040066,
            0.40721626785040044,
            0.16675952108434378,
            0.16675952108434364,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            0.4353002296391376,
            0.43530022963913784,
            0.1784444372726958,
            0.17844443727269582,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            0.4297181206523321,
            0.4297181206523325,
            0.17683059662141104,
            0.17683059662141118,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            0.4563061538097892,
            0.45630615380978823,
            0.18896605372769398,
            0.18896605372769357,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            0.42868186647283657,
            0.4286818664728357,
            0.17732426650078859,
            0.17732426650078814,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            0.4363283710194509,
            0.43632837101945204,
            0.18073875330979583,
            0.18073875330979633,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            0.4164016692781507,
            0.41640166927815186,
            0.17120274485655348,
            0.171202744856554,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            0.3956722333918168,
            0.3956722333918169,
            0.16185382682093818,
            0.1618538268209383,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            0.40074742225209253,
            0.4007474222520928,
            0.1637175794985678,
            0.16371757949856788,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            0.380483029250028,
            0.3804830292500278,
            0.15515437464176265,
            0.15515437464176265,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            -0.2388264375398923,
            -0.2388264375398914,
            0.14532595961658287,
            0.14532595961658334,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            -0.008572276672876411,
            -0.008572276672875134,
            0.1774866649515709,
            0.17748666495157142,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            0.21827280545962327,
            0.21827280545962424,
            0.2220135357854755,
            0.222013535785476,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            0.31418188572614525,
            0.31418188572614564,
            0.22195820948818912,
            0.22195820948818934,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            0.4476237091386184,
            0.4476237091386175,
            0.24724451707518497,
            0.2472445170751845,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            0.4565039584664099,
            0.45650395846640895,
            0.22745711249685308,
            0.22745711249685266,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            0.5503851129624154,
            0.5503851129624173,
            0.2503917100241949,
            0.2503917100241957,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            0.507350612150201,
            0.5073506121502019,
            0.21934324414922365,
            0.21934324414922415,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            0.5562486047109594,
            0.5562486047109596,
            0.2316485191114935,
            0.23164851911149356,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            0.48383778639655683,
            0.4838377863965566,
            0.194730821427455,
            0.19473082142745496,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            0.5067959807807199,
            0.506795980780722,
            0.20113715240745478,
            0.2011371524074559,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            0.42655573754293186,
            0.4265557375429339,
            0.16417597251024682,
            0.16417597251024788,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            0.4185939963363059,
            0.4185939963363061,
            0.15952683122300343,
            0.15952683122300346,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            0.34041140640754475,
            0.3404114064075435,
            0.12547707652609247,
            0.12547707652609205,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            0.33130911220195647,
            0.33130911220195614,
            0.12197341166486306,
            0.12197341166486311,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            0.26405379779377447,
            0.2640537977937747,
            0.09318841840402961,
            0.0931884184040298,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            0.25071170671186394,
            0.2507117067118643,
            0.08875608691136452,
            0.08875608691136472,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            0.19598108047305396,
            0.19598108047305401,
            0.06546898672294421,
            0.06546898672294431,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            0.17014538308114796,
            0.17014538308114918,
            0.05544124100230613,
            0.0554412410023066,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            0.14234869790555654,
            0.14234869790555757,
            0.044245450305331835,
            0.044245450305332244,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            -0.5091467481118493,
            -0.5091467481118483,
            0.017941506342799257,
            0.017941506342799618,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            -0.2740893971683367,
            -0.27408939716833514,
            0.052241617289731646,
            0.052241617289732006,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            -0.10354868305191589,
            -0.10354868305191428,
            0.07222601607723342,
            0.07222601607723408,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            0.01661846172941467,
            0.016618461729414807,
            0.08145524988015601,
            0.08145524988015608,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            0.09795151148394185,
            0.09795151148394297,
            0.08283997695192209,
            0.08283997695192256,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            0.14991670005762325,
            0.14991670005762409,
            0.07879139715693627,
            0.07879139715693662,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            0.18037636242197974,
            0.1803763624219798,
            0.07136113382288911,
            0.07136113382288908,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            0.19595043762863176,
            0.19595043762863226,
            0.062345815873464216,
            0.06234581587346444,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            0.20228029801426928,
            0.20228029801426872,
            0.05335802567312438,
            0.053358025673124185,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            0.2042370167172295,
            0.20423701671722905,
            0.04588091889282403,
            0.04588091889282387,
            0.0,
            0.0,
        ],
    ]
end

sleep(0.1) # used for good printing
println("End test")

end # End Main Module 
